/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#include <stdlib.h>

#include <mimimi.h>

void mimimi_camera_stamp(struct mimimi_engine *engine, struct mimimi_position *camera, int x, int y, int z, void *texture)
{
	z += 8;
	if (z <= 0) return;
	
	x -= camera->x;
	x = mimimi_div_up(x, z);
	x += engine->size.width / 2;
	
	y -= camera->y;
	y = mimimi_div_up(y, z);
	y += engine->size.height / 2;
	
	(*engine->stamp)(engine->data, x, y, texture);
}

void mimimi_display_tick(struct mimimi_display *display, struct mimimi_engine *engine, struct mimimi_sprite *sprite)
{
	struct mimimi_appearance *appearance;
	struct mimimi_position *position;
	struct mimimi_physics *physics;
	struct mimimi_life *life;
	struct mimimi_walk *walk;
	int dx, dy;
	struct mimimi_video_set *videos;
	struct mimimi_video *video;
	void *texture;
	int x, y;
	
	appearance = &display->appearance;
	
	position = &sprite->position;
	physics = &sprite->physics;
	life = &sprite->life;
	walk = &sprite->walk;
	
	x = display->width * -8;
	y = display->height * -8;
	
	if (life->knocked_time == 0 && walk->direction != 0)
		display->direction = walk->direction;
	
	dx = physics->dx;
	dy = physics->dy;
	videos = &appearance->right;
	
	if (display->direction == 1)
	{
		videos = &appearance->left;
		dx *= -1;
	}
	
	if (dx > 8191) dx = 8191;
	if (dx < 0) dx = 0;
	
	if (life->knocked_time == 0)
	{
		if (physics->airborne != 0)
		{
			if (dy < -8191) dy = -8191;
			if (dy > 8191) dy = 8191;
			dy += 8191;
			
			video = videos->jumping + dx * videos->jumping_video_count / 8192;
			texture = video->textures[dy * video->count / 16384];
			
			mimimi_camera_stamp(engine, display->camera, position->x + x, position->y + y, 0, texture);
			return;
		}
		
		if (physics->airborne != 0)
		{
			texture = videos->jumping->textures[dx * videos->jumping->count / 8192];
			mimimi_camera_stamp(engine, display->camera, position->x + x, position->y + y, 0, texture);
			return;
		}
		
		video = videos->standing + dx * videos->standing_video_count / 8192;
		display->animation_time += dx / 1024;
	}
	else
	{
		if (physics->airborne == 0)
			video = videos->knocked,
			y += 212;
		else
			video = videos->falling,
			y += 128;
		
		if (dy < 0) dy = 0;
		display->animation_time += dy / 2048;
	}
	
	/* do not assume 'char' is eight bits */
	display->animation_time &= 0xFF;
	
	texture = video->textures[display->animation_time * video->count / 256];
	mimimi_camera_stamp(engine, display->camera, position->x + x, position->y + y, 0, texture);
}

void mimimi_background_tick(struct mimimi_background *background, struct mimimi_engine *engine, struct mimimi_position *camera, struct mimimi_position *offset)
{
	mimimi_camera_stamp(engine, camera, background->x + offset->x, background->y + offset->y, background->z, background->texture);
}

static void mimimi_populate_video_set(struct mimimi_video_set *set)
{
	static int video_count = 8;
	static int image_count = 8;
	static int width = 48;
	static int height = 48;
	
	int i, j, k;
	struct mimimi_video *videos, *standing, *knocked, *falling, *jumping;
	struct mimimi_image *images, *image;
	unsigned char *colors;
	void **textures;
	
	videos = malloc(sizeof *videos * (video_count * 2 + 2));
	images = malloc(sizeof *images * (image_count * video_count * 2 + 2));
	textures = malloc(sizeof *textures * (image_count * video_count * 2 + 2));
	colors = malloc((image_count * video_count * 2 + 2) * 48 * 48);
	
	if (videos == NULL) exit(1);
	if (images == NULL) exit(1);
	if (textures == NULL) exit(1);
	if (colors == NULL) exit(1);
	
	set->standing = videos;
	set->standing_video_count = video_count;
	videos += video_count;
	
	set->jumping = videos;
	set->jumping_video_count = video_count;
	videos += video_count;
	
	for (i = 0 ; i < video_count ; i++)
	{
		standing = set->standing + i;
		standing->count = image_count;
		standing->images = images;
		images += image_count;
		standing->textures = textures;
		textures += image_count;
		
		for (j = 0 ; j < image_count ; j++)
		{
			image = standing->images + j;
			image->width = width;
			image->height = height;
			image->colors = colors;
			colors += width * height;
			for (k = 0 ; k < width * height ; k++) image->colors[k] = 0;
		}
		
		jumping = set->jumping + i;
		jumping->count = image_count;
		jumping->images = images;
		images += image_count;
		jumping->textures = textures;
		textures += image_count;
		
		for (j = 0 ; j < image_count ; j++)
		{
			image = jumping->images + j;
			image->width = width;
			image->height = height;
			image->colors = colors;
			colors += width * height;
			for (k = 0 ; k < width * height ; k++) image->colors[k] = 0;
		}
	}
	
	knocked = videos++;
	knocked->count = 1;
	knocked->images = images++;
	knocked->images[0].width = width;
	knocked->images[0].height = height;
	knocked->images[0].colors = colors;
	knocked->textures = textures++;
	colors += width * height;
	for (i = 0 ; i < width * height ; i++) knocked->images[0].colors[i] = 0;
	
	falling = videos++;
	falling->count = 1;
	falling->images = images++;
	falling->images[0].width = width;
	falling->images[0].height = height;
	falling->images[0].colors = colors;
	colors += width * height;
	falling->textures = textures++;
	for (i = 0 ; i < width * height ; i++) falling->images[0].colors[i] = 0;
	
	set->knocked = knocked;
	set->falling = falling;
}

static void mimimi_populate_appearance(struct mimimi_display *display, struct mimimi_model *model, struct mimimi_engine *engine)
{
	mimimi_populate_video_set(&display->appearance.left);
	mimimi_populate_video_set(&display->appearance.right);
	
	mimimi_appearance(&display->appearance, model, 24, 37);
	mimimi_prepare_appearance(&display->appearance, engine);
}

void mimimi_display(struct mimimi_display *display, struct mimimi_model *model, struct mimimi_engine *engine)
{
	display->animation_time = 0;
	display->direction = 1;
	display->width = 24;
	display->height = 48;
	mimimi_populate_appearance(display, model, engine);
}
