/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#ifndef MIMIMI_H
#define MIMIMI_H

struct mimimi_position
{
	int x, y;
};

struct mimimi_size
{
	int width, height;
};

struct mimimi_image
{
	int width, height;
	unsigned char *colors;
};

struct mimimi_engine
{
	void *data;
	void *(*texture)(void *data, struct mimimi_image *image);
	void (*invalidate)(void *data, void *texture);
	void (*stamp)(void *data, int x, int y, void *texture);
	struct mimimi_size size;
};

struct mimimi_physics
{
	int dx, dy;
	unsigned char airborne;
	int width, height;
	int gravity;
};

struct mimimi_life
{
	int knocked_time;
	int immunity_time;
	int pristinity;
	int max_pristinity;
	int recovery_time;
};

struct mimimi_fall
{
	int max_speed;
	int knocked_time;
};

struct mimimi_walk
{
	int ground_speed;
	int airborne_speed;
	int direction;
};

struct mimimi_ground
{
	int width, height;
	unsigned char tiles[0x2000];
};

struct mimimi_sprite
{
	struct mimimi_position position;
	struct mimimi_physics physics;
	struct mimimi_life life;
	struct mimimi_fall fall;
	struct mimimi_walk walk;
	struct mimimi_ground *ground;
	struct mimimi_position offset;
};

struct mimimi_conveyor
{
	struct mimimi_sprite *sprite;
	struct mimimi_ground *ground;
	unsigned char was_airborne;
	int speed;
};

struct mimimi_positioned_physics
{
	struct mimimi_physics *physics;
	struct mimimi_physics other;
	struct mimimi_position transform;
	struct mimimi_position *position;
	struct mimimi_position previous;
	unsigned char was_airborne;
};

struct mimimi_trampoline
{
	struct mimimi_sprite *sprite;
	struct mimimi_ground *ground;
	int restitution;
	int dy;
};

struct mimimi_video
{
	int count;
	struct mimimi_image *images;
	void **textures;
};

struct mimimi_video_set
{
	int standing_video_count;
	int jumping_video_count;
	struct mimimi_video *standing;
	struct mimimi_video *knocked;
	struct mimimi_video *falling;
	struct mimimi_video *jumping;
};

struct mimimi_appearance
{
	struct mimimi_video_set left, right;
};

struct mimimi_font
{
	unsigned char glyphs[0x60][0x10];
};

struct mimimi_pose_layer
{
	int slant;
	int y_angle;
	int z_angle;
};

struct mimimi_pose
{
	int x;
	int y;
	int count;
	struct mimimi_pose_layer layers[0x100];
};

struct mimimi_movement
{
	int count;
	struct mimimi_pose *poses;
};

struct mimimi_background
{
	void *texture;
	int x, y, z;
};

struct mimimi_display
{
	struct mimimi_position *camera;
	struct mimimi_appearance appearance;
	unsigned char animation_time;
	int direction;
	int width, height;
};

struct mimimi_row
{
	int width;
	int height;
	int count;
	unsigned char colors[0x80];
};

struct mimimi_layer
{
	int x, y, z;
	int width;
	int amplifier;
	int count;
	struct mimimi_row rows[0x80];
};

struct mimimi_model_layer
{
	int parent_index;
	struct mimimi_layer *layer;
};

struct mimimi_model
{
	int count;
	struct mimimi_model_layer layers[0x80];
};

struct mimimi_history
{
	unsigned int left, right;
};

struct mimimi_clamped_camera
{
	struct mimimi_position *input;
	struct mimimi_position output;
	struct mimimi_position offset;
	struct mimimi_size size;
};

struct mimimi_dialogue_paragraph
{
	struct mimimi_model *model;
	char *name;
	char *text;
};

struct mimimi_dialogue
{
	struct mimimi_dialogue_paragraph *paragraphs;
	int count;
};

struct mimimi_strip_dialogue
{
	struct mimimi_engine *engine;
	int count;
	void *texture;
	int notches[0x80];
	unsigned char held;
	int i;
	int x, y;
	struct mimimi_image image;
	unsigned char colors[128 * 2048];
};

struct mimimi_toast_dialogue
{
	struct mimimi_engine *engine;
	int count;
	void *textures[0x80];
	int delays[0x80];
	int heights[0x80];
	unsigned char colors[128 * 2048];
	struct mimimi_image images[128];
	int i, j, k;
};

struct mimimi_momentum
{
	int x;
	int dx;
};

struct mimimi_platform
{
	struct mimimi_positioned_physics physics;
	struct mimimi_ground *ground;
	void *texture;
};

struct mimimi_linear_motion
{
	int *value;
	int origin;
	int scale;
};

struct mimimi_angular_motion
{
	struct mimimi_position *value;
	struct mimimi_position origin;
	int angle;
	int scale;
	int height;
};

struct mimimi_logo_line
{
	unsigned char *packed;
	struct mimimi_size size;
	struct mimimi_position position;
};

struct mimimi_logo
{
	struct mimimi_logo_line *lines;
	unsigned char colors[8][16384];
	struct mimimi_image images[8];
	void *textures[8];
	int y_offsets[8];
	int count;
};

struct mimimi_platform_set
{
	struct mimimi_momentum momentum;
	int balance;
	int strength;
	int linear_weight;
	int angular_weight;
	int min, max;
	
	int linear_platform_count;
	struct mimimi_platform *linear_platforms;
	struct mimimi_linear_motion *linear_motions;
	
	int angular_platform_count;
	struct mimimi_platform *angular_platforms;
	struct mimimi_angular_motion *angular_motions;
};

struct mimimi_area
{
	struct mimimi_ground *ground;
	struct mimimi_position offset;
	struct mimimi_background *backgrounds;
	int background_count;
	struct mimimi_background *foregrounds;
	int foreground_count;
};

struct mimimi_stage
{
	struct mimimi_display **displays;
	struct mimimi_sprite *sprites;
	int sprite_count;
	struct mimimi_position camera;
	struct mimimi_clamped_camera clamped_camera;
	struct mimimi_area *areas;
	int area_count;
	struct mimimi_platform_set *platform_sets;
	int platform_set_count;
};

extern struct mimimi_font *mimimi_font;

extern struct mimimi_model *mimimi_mango;
extern struct mimimi_model *mimimi_pepper;

extern struct mimimi_model *mimimi_homura_flash;
extern struct mimimi_model *mimimi_mirai_flash;
extern struct mimimi_model *mimimi_mirai_yang;
extern struct mimimi_model *mimimi_haruka_flash;

extern struct mimimi_model *mimimi_ryoubi_flash;
extern struct mimimi_model *mimimi_ryouna_flash;

extern signed char *mimimi_sine;
extern signed char *mimimi_cosine;

extern struct mimimi_logo_line *mimimi_logo_lines;
extern int mimimi_logo_line_count;

int mimimi_div(int n, int q);
int mimimi_div_up(int n, int q);
int mimimi_div_down(int n, int q);
int mimimi_mod(int n, int q);
int mimimi_angle(int x, int y);

void mimimi_camera_tick(struct mimimi_position *camera, struct mimimi_position *position, int ox, int oy);
void mimimi_camera_stamp(struct mimimi_engine *engine, struct mimimi_position *camera, int x, int y, int z, void *texture);
void mimimi_clamped_camera_tick(struct mimimi_clamped_camera *camera, int width, int height);

void mimimi_collision_physics_tick(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground);
void mimimi_dynamics_tick(struct mimimi_physics *physics, struct mimimi_position *position);
void mimimi_physics_tick(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground);
void mimimi_physics(struct mimimi_physics *physics, int width, int height);

void mimimi_life_malloctick(struct mimimi_life *life);
void mimimi_fall_tick(struct mimimi_fall *fall, struct mimimi_physics *physics, struct mimimi_life *life);
void mimimi_walk_tick(struct mimimi_walk *walk, struct mimimi_physics *physics);

void mimimi_sprite(struct mimimi_sprite *sprite, struct mimimi_ground *ground, int x, int y, int width, int height);
void mimimi_sprite_tick(struct mimimi_sprite *sprite);

void mimimi_conveyor_tick(struct mimimi_conveyor *conveyor);
void mimimi_positioned_physics_tick(struct mimimi_positioned_physics *physics, struct mimimi_ground *ground);
void mimimi_positioned_physics(struct mimimi_positioned_physics *physics, struct mimimi_physics *other, struct mimimi_position *position);
void mimimi_trampoline_tick(struct mimimi_trampoline *trampoline);

void mimimi_appearance(struct mimimi_appearance *appearance, struct mimimi_model *model, int x, int y);
void mimimi_video(struct mimimi_video *video, struct mimimi_engine *engine);
void mimimi_invalidate_video(struct mimimi_video *video, struct mimimi_engine *engine);
void mimimi_prepare_appearance(struct mimimi_appearance *appearance, struct mimimi_engine *engine);
void mimimi_invalidate_appearance(struct mimimi_appearance *appearance, struct mimimi_engine *engine);

int mimimi_measure_character(struct mimimi_font *font, char ch);
int mimimi_draw_character(struct mimimi_font *font, char ch, struct mimimi_image *target, int x, int y, unsigned char color);

int mimimi_measure_segment(struct mimimi_font *font, char *text, int count);
int mimimi_draw_segment(struct mimimi_font *font, char *text, int count, struct mimimi_image *target, int x, int y, unsigned char color);

char *mimimi_skip_word(char *text);
int mimimi_count_word(char *text);
int mimimi_measure_word(struct mimimi_font *font, char *text);
int mimimi_draw_word(struct mimimi_font *font, char *text, struct mimimi_image *target, int x, int y, unsigned char color);

char *mimimi_skip_paragraph(char *text);
int mimimi_count_paragraph(char *text);
int mimimi_measure_paragraph(struct mimimi_font *font, char *text);
int mimimi_draw_paragraph(struct mimimi_font *font, char *text, struct mimimi_image *target, int x, int y, unsigned char color);

char *mimimi_skip_line(struct mimimi_font *font, char *text, int width);
int mimimi_count_line(struct mimimi_font *font, char *text, int width);
int mimimi_measure_line(struct mimimi_font *font, char *text, int width);
int mimimi_draw_line(struct mimimi_font *font, char *text, int width, struct mimimi_image *target, int x, int y, unsigned char color);

int mimimi_measure_text(struct mimimi_font *font, char *text, int width, int line_height);
int mimimi_draw_text(struct mimimi_font *font, char *text, int width, int line_height, struct mimimi_image *target, int x, int y, unsigned char color);

void mimimi_movement(struct mimimi_video *video, struct mimimi_model *model, struct mimimi_movement *movement);
void mimimi_pose(struct mimimi_image *image, struct mimimi_model *model, struct mimimi_pose *pose);

void mimimi_background_tick(struct mimimi_background *background, struct mimimi_engine *engine, struct mimimi_position *camera, struct mimimi_position *offset);
void mimimi_display_tick(struct mimimi_display *display, struct mimimi_engine *engine, struct mimimi_sprite *sprite);

unsigned char mimimi_ground_tile(struct mimimi_ground *ground, int x, int y);

void mimimi_history_tick(struct mimimi_history *history, unsigned char left, unsigned char right);
void mimimi_controls_tick(struct mimimi_walk *walk, struct mimimi_history *history, void (*jump)(void *data), void *data);
void mimimi_jump(void *data);

void mimimi_strip_dialogue(struct mimimi_strip_dialogue *strip_dialogue, struct mimimi_dialogue *dialogue, struct mimimi_engine *engine);
unsigned char mimimi_strip_dialogue_tick(struct mimimi_strip_dialogue *dialogue, unsigned char held);

void mimimi_toast_dialogue(struct mimimi_toast_dialogue *toast_dialogue, struct mimimi_dialogue *dialogue, struct mimimi_engine *engine);
unsigned char mimimi_toast_dialogue_top_tick(struct mimimi_toast_dialogue *dialogue);
unsigned char mimimi_toast_dialogue_bottom_tick(struct mimimi_toast_dialogue *dialogue);

void mimimi_stage(struct mimimi_stage *stage, struct mimimi_ground **grounds, int count);
void mimimi_stage_tick(struct mimimi_stage *stage, struct mimimi_engine *engine);
struct mimimi_sprite *mimimi_spawn(struct mimimi_stage *stage, struct mimimi_display *display, int x, int y, int width, int height);

void mimimi_linear_propagate(struct mimimi_momentum *momentum, int value, int x0, int scale);
void mimimi_linear_apply(struct mimimi_momentum *momentum, int *value, int x0, int scale);

void mimimi_angular_propagate(struct mimimi_momentum *momentum, struct mimimi_position *position, int x0, int y0, int a0, int scale);
void mimimi_angular_apply(struct mimimi_momentum *momentum, struct mimimi_position *position, int x0, int y0, int a0, int scale, int length);

void mimimi_linear_motion_propagate(struct mimimi_linear_motion *motion, struct mimimi_momentum *momentum);
void mimimi_linear_motion_apply(struct mimimi_linear_motion *motion, struct mimimi_momentum *momentum);

void mimimi_angular_motion_propagate(struct mimimi_angular_motion *motion, struct mimimi_momentum *momentum);
void mimimi_angular_motion_apply(struct mimimi_angular_motion *motion, struct mimimi_momentum *momentum);

void mimimi_platform_display_tick(struct mimimi_platform *platform, struct mimimi_position *camera, int x, int y, struct mimimi_engine *engine);
struct mimimi_platform_set *mimimi_add_platform_set(struct mimimi_stage *stage);
void mimimi_spawn_linear_platform(struct mimimi_stage *stage, struct mimimi_platform_set *platforms, struct mimimi_ground *ground, void *texture, int x, int y, int scale);
void mimimi_spawn_angular_platform(struct mimimi_stage *stage, struct mimimi_platform_set *platforms, struct mimimi_ground *ground, void *texture, int x, int y, int a, int scale, int height);

void mimimi_platform_set_tick(struct mimimi_platform_set *platform_set);
void mimimi_platform_set_display_tick(struct mimimi_platform_set *platforms, struct mimimi_engine *engine, struct mimimi_position *camera);

void mimimi_logo(struct mimimi_engine *engine, struct mimimi_logo *logo, struct mimimi_logo_line *lines, int count);
unsigned char mimimi_logo_tick(struct mimimi_engine *engine, struct mimimi_logo *logo, int x, int y);

void mimimi_unpack_bits(int size, unsigned char *unpacked, unsigned char *packed);

void mimimi_linear_motion(struct mimimi_linear_motion *motion, int *value, int origin, int scale);
void mimimi_angular_motion(struct mimimi_angular_motion *motion, struct mimimi_position *value, int x0, int y0, int a0, int scale, int height);

void mimimi_display(struct mimimi_display *display, struct mimimi_model *model, struct mimimi_engine *engine);

void mimimi_background_image(struct mimimi_image *image, struct mimimi_ground *ground, unsigned char *colors);
void *mimimi_background(struct mimimi_engine *engine, struct mimimi_ground *ground, unsigned char *colors);

int mimimi_wasm(void *(*start)(struct mimimi_engine *engine), void (*tick)(void *chapter, unsigned char left, unsigned char right));
int mimimi_fbdev(void *(*start)(struct mimimi_engine *engine), void (*tick)(void *chapter, unsigned char left, unsigned char right));
int mimimi_minifb(void *(*start)(struct mimimi_engine *engine), void (*tick)(void *chapter, unsigned char left, unsigned char right));
int mimimi_sdl2(void *(*start)(struct mimimi_engine *engine), void (*tick)(void *chapter, unsigned char left, unsigned char right));

#endif
